Imports System
Imports System.Xml.Serialization
<Serializable()> Public Class calculations
    '/////////////////////////////////////////////////////////
    '  Based on the calculations used by the
    '  National Oceanic and Atmospheric Administration Surface Radiation Research Branch:
    '  http://www.srrb.noaa.gov/highlights/sunrise/sunrise.html
    '  which contains a link to the formulas used.
    '  Created on:      12-Jan-2004 05:11:41 PM
    '  Modified on:     02/08/2004 
    '  Original author: Ken Arway
    '  Conversion to VB: Ken Cox (kjopc@hotmail.com) 
    '  Conversion tool: http://www.gotdotnet.com/Community/UserSamples/Details.aspx?SampleGuid=4006c1b7-ec75-4ede-a1eb-01dcd7d4d651 
    '  This version uses Toronto, Ontario, Canada as the default coordinates
    '/////////////////////////////////////////////////////////

    Private _location As String

    Public Property location() As String
        Get
            Return _location
        End Get
        Set(ByVal Value As String)
            _location = Value
        End Set
    End Property

    Private _latitude As Double = 43.692  ' in decimal degrees, positive for North latitudes
    Public Property latitude() As Double
        Get
            Return _latitude
        End Get
        Set(ByVal Value As Double)
            _latitude = Value
        End Set
    End Property


    Private _longitude As Double = 79.3716 ' in decimal degrees, positive for West longitudes
    Public Property longitude() As Double
        Get
            Return _longitude
        End Get
        Set(ByVal Value As Double)
            _longitude = Value
        End Set
    End Property

    Private _timezone As Double = 5 ' hours offset from UTC, positive for West longitudes    
    Public Property timezone() As Double
        Get
            Return _timezone
        End Get
        Set(ByVal Value As Double)
            _timezone = Value
        End Set
    End Property

    Private _dst As Integer = 0 ' Daylight Savings Time in effect:  1 = yes, 0 = no
    Public Property dst() As Integer
        Get
            Return _dst
        End Get
        Set(ByVal Value As Integer)
            _dst = Value
        End Set
    End Property

    Private _nowtimeLocal As DateTime
    Public Property nowtimeLocal() As DateTime
        Get
            Return _nowtimeLocal
        End Get
        Set(ByVal Value As DateTime)
            _nowtimeLocal = Value
        End Set
    End Property

    Private _nowtime As DateTime
    Public Property nowtime() As DateTime
        Get
            Return _nowtime
        End Get
        Set(ByVal Value As DateTime)
            _nowtime = Value
        End Set
    End Property

    Private _nowtimeUTC As DateTime
    Public Property nowtimeUTC() As DateTime
        Get
            Return _nowtimeUTC
        End Get
        Set(ByVal Value As DateTime)
            _nowtimeUTC = Value
        End Set
    End Property


    Private _nowtimeZ As DateTime
    Public Property nowtimeZ() As DateTime
        Get
            Return _nowtimeZ
        End Get
        Set(ByVal Value As DateTime)
            _nowtimeZ = Value
        End Set
    End Property


    Private _solarnoonUTC As DateTime
    Public Property solarnoonUTC() As DateTime
        Get
            Return _solarnoonUTC
        End Get
        Set(ByVal Value As DateTime)
            _solarnoonUTC = Value
        End Set
    End Property


    Private _zenith As Double
    Public Property zenith() As Double
        Get
            Return _zenith
        End Get
        Set(ByVal Value As Double)
            _zenith = Value
        End Set
    End Property


    Private _zenithvalue As String
    Public Property zenithvalue() As String
        Get
            Return _zenithvalue
        End Get
        Set(ByVal Value As String)
            _zenithvalue = Value
        End Set
    End Property


    Private _doy As Integer
    Public Property doy() As Integer
        Get
            Return _doy
        End Get
        Set(ByVal Value As Integer)
            _doy = Value
        End Set
    End Property

    Private _year As Integer
    Public Property year() As Integer
        Get
            Return _year
        End Get
        Set(ByVal Value As Integer)
            _year = Value
        End Set
    End Property


    Private _month As Integer
    Public Property month() As Integer
        Get
            Return _month
        End Get
        Set(ByVal Value As Integer)
            _month = Value
        End Set
    End Property


    Private _day As Integer
    Public Property day() As Integer
        Get
            Return _day
        End Get
        Set(ByVal Value As Integer)
            _day = Value
        End Set
    End Property

    Private _sunriseLocal As DateTime
    Public Property sunriseLocal() As DateTime
        Get
            Return _sunriseLocal
        End Get
        Set(ByVal Value As DateTime)
            _sunriseLocal = Value
        End Set
    End Property

    Private _sunriseUTC As DateTime
    Public Property sunriseUTC() As DateTime
        Get
            Return _sunriseUTC
        End Get
        Set(ByVal Value As DateTime)
            _sunriseUTC = Value
        End Set
    End Property

    Private _sunsetLocal As DateTime
    Public Property sunsetLocal() As DateTime
        Get
            Return _sunsetLocal
        End Get
        Set(ByVal Value As DateTime)
            _sunsetLocal = Value
        End Set
    End Property

    Private _sunsetUTC As DateTime
    Public Property sunsetUTC() As DateTime
        Get
            Return _sunsetUTC
        End Get
        Set(ByVal Value As DateTime)
            _sunsetUTC = Value
        End Set
    End Property

    Private _hoursOfDaylight As TimeSpan
    Public Property hoursOfDaylight() As TimeSpan
        Get
            Return _hoursOfDaylight
        End Get
        Set(ByVal Value As TimeSpan)
            _hoursOfDaylight = Value
        End Set
    End Property

End Class

Public Class SunRoutines

    'Official (O), Civil (C), Nautical (N) or Astronimical (A) horizon limit
    '/ <summary>
    '/ The entry point for the program.
    '/ </summary>
    Public Function GetSunriseSunset(ByVal location As String, _
    ByVal latitude As Double, _
    ByVal longitude As Double, _
    ByVal timezone As Double, _
    ByVal dst As Integer, _
    ByVal zenithvalue As String) As calculations

        Dim sr As New calculations
        sr.location = location
        sr.latitude = latitude
        sr.longitude = longitude
        sr.timezone = timezone
        sr.dst = dst
        sr.zenithvalue = zenithvalue
        sr.nowtimeLocal = DateTime.Now
        sr.nowtimeZ = sr.nowtimeLocal.Date
        sr.nowtimeUTC = sr.nowtimeLocal.ToUniversalTime()
        sr.nowtime = sr.nowtimeUTC.AddHours(-sr.timezone + sr.dst)
        sr.doy = sr.nowtime.DayOfYear
        sr.day = sr.nowtime.Day
        sr.month = sr.nowtime.Month
        sr.year = sr.nowtime.Year

        If sr.latitude >= -90 AndAlso sr.latitude < -89.7 Then
            '     sb.append("All latitudes between 89.7 and 90 S" & "<br />" & " will be set to -89.7")
            sr.latitude = -89.7
        End If
        If sr.latitude <= 90 AndAlso sr.latitude > 89.7 Then
            '    sb.append("All latitudes between 89.7 and 90 N" & "<br />" & " will be set to 89.7")
            sr.latitude = 89.7
        End If

        Dim JD As Double = calcJD(sr.day, sr.month, sr.year)
        Dim t As Double = calcTimeJulianCent(JD)
        Dim eqtime As Double = calcEquationOfTime(t)
        Dim solarDec As Double = calcSunDeclination(t)
        sr.solarnoonUTC = sr.nowtimeZ.AddMinutes(calcSolNoonUTC(t, sr.longitude))
        doCalc(JD, t, eqtime, solarDec, sr.zenithvalue, sr)
        Return sr
    End Function 'Main

    '/ <summary>
    '/ The main calculation starting point and results display.
    '/ </summary>
    '/ <param name='JD'>Julian Day</param>
    '/ <param name='t'>Julian centuries since J2000.0</param>
    '/ <param name='eqtime'>Equation of Time</param>
    '/ <param name='solarDec'>Solar Declination</param>
    '/ <param name='zenithValue'>Sets the horizon limit</param>
    Public Function doCalc(ByVal JD As Double, ByVal t As Double, ByVal eqtime As Double, ByVal solarDec As Double, ByVal zenithValue As String, ByRef sr As calculations) As String
        Dim nosunrise As Boolean = False
        Dim nosunset As Boolean = False

        Select Case zenithValue
            Case "O"
                sr.zenith = 90.8333
            Case "C"
                sr.zenith = 96.0
            Case "N"
                sr.zenith = 102.0
            Case "A"
                sr.zenith = 108.0
        End Select

        ' Does the calculations.
        Try
            ' test for sunrise.
            Dim sunriseUTC As DateTime = sr.nowtimeZ.AddMinutes(calcSunriseUTC(sr.zenith, JD, sr.latitude, sr.longitude))
        Catch exc As Exception
            nosunrise = True
        End Try
        Try
            ' test for sunset.
            Dim sunsetUTC As DateTime = sr.nowtimeZ.AddMinutes(calcSunsetUTC(sr.zenith, JD, sr.latitude, sr.longitude))
        Catch exc As Exception
            nosunset = True
        End Try

        If nosunrise = False AndAlso nosunset = False Then
            ' there is both a sunrise and sunset.
            sr.sunriseUTC = sr.nowtimeZ.AddMinutes(calcSunriseUTC(sr.zenith, JD, sr.latitude, sr.longitude))
            sr.sunriseLocal = sr.sunriseUTC.AddHours(-sr.timezone + sr.dst)
            ' round to nearest minute.
            sr.sunriseLocal = IIf(sr.sunriseLocal.Second >= 30, sr.sunriseLocal.AddMinutes(1).AddSeconds(-sr.sunriseLocal.Second).AddMilliseconds(-sr.sunriseLocal.Millisecond), sr.sunriseLocal.AddSeconds(-sr.sunriseLocal.Second).AddMilliseconds(-sr.sunriseLocal.Millisecond))     'TODO: For performance reasons this should be changed to nested IF statements
            sr.sunsetUTC = sr.nowtimeZ.AddMinutes(calcSunsetUTC(sr.zenith, JD, sr.latitude, sr.longitude))
            sr.sunsetLocal = sr.sunsetUTC.AddHours(-sr.timezone + sr.dst)
            ' round to nearest minute.
            sr.sunsetLocal = IIf(sr.sunsetLocal.Second >= 30, sr.sunsetLocal.AddMinutes(1).AddSeconds(-sr.sunsetLocal.Second).AddMilliseconds(-sr.sunsetLocal.Millisecond), sr.sunsetLocal.AddSeconds(-sr.sunsetLocal.Second).AddMilliseconds(-sr.sunsetLocal.Millisecond))     'TODO: For performance reasons this should be changed to nested IF statements
            sr.hoursOfDaylight = sr.sunsetLocal.Subtract(sr.sunriseLocal)
            ' Formats and displays the results.
            sr.solarnoonUTC = sr.solarnoonUTC.AddHours((-sr.timezone + sr.dst))
        Else ' there is either no sunrise or no sunset.
            If nosunrise = True Then
                ' there is no sunrise
                ' if Northern hemisphere and spring or summer, OR
                ' if Southern hemisphere and fall or winter, use
                ' previous sunrise and next sunset
                If sr.latitude > 66.4 AndAlso sr.doy > 79 AndAlso sr.doy < 267 OrElse (sr.latitude < -66.4 AndAlso (sr.doy < 83 OrElse sr.doy > 263)) Then
                    Dim newjd As Double = findRecentSunrise(JD, sr.latitude, sr.longitude, zenithValue, sr)

                    Dim newSunrise As DateTime = calcDateFromJD(newjd)
                    sr.sunriseUTC = newSunrise.AddMinutes(calcSunriseUTC(sr.zenith, newjd, sr.latitude, sr.longitude))
                    sr.sunriseLocal = sr.sunriseUTC.AddHours(-sr.timezone + sr.dst)
                    ' round to nearest minute.
                    sr.sunriseLocal = IIf(sr.sunriseLocal.Second >= 30, sr.sunriseLocal.AddMinutes(1).AddSeconds(-sr.sunriseLocal.Second).AddMilliseconds(-sr.sunriseLocal.Millisecond), sr.sunriseLocal.AddSeconds(-sr.sunriseLocal.Second).AddMilliseconds(-sr.sunriseLocal.Millisecond))
                    'TODO: For performance reasons this should be changed to nested IF statements


                    ' if Northern hemisphere and fall or winter, OR
                    ' if Southern hemisphere and spring or summer, use
                    ' next sunrise and previous sunset
                ElseIf sr.latitude > 66.4 AndAlso (sr.doy < 83 OrElse sr.doy > 263) OrElse (sr.latitude < -66.4 AndAlso sr.doy > 79 AndAlso sr.doy < 267) Then
                    Dim newjd As Double = findNextSunrise(JD, sr.latitude, sr.longitude, sr.zenithvalue, sr)

                    Dim newSunrise As DateTime = calcDateFromJD(newjd)
                    sr.sunriseUTC = newSunrise.AddMinutes(calcSunriseUTC(sr.zenith, newjd, sr.latitude, sr.longitude))
                    sr.sunriseLocal = sr.sunriseUTC.AddHours(-sr.timezone + sr.dst)
                    ' round to nearest minute.
                    sr.sunriseLocal = IIf(sr.sunriseLocal.Second >= 30, sr.sunriseLocal.AddMinutes(1).AddSeconds(-sr.sunriseLocal.Second).AddMilliseconds(-sr.sunriseLocal.Millisecond), sr.sunriseLocal.AddSeconds(-sr.sunriseLocal.Second).AddMilliseconds(-sr.sunriseLocal.Millisecond))     'TODO: For performance reasons this should be changed to nested IF statements

                End If
            Else ' there is a sunrise.
                sr.sunriseUTC = sr.nowtimeZ.AddMinutes(calcSunriseUTC(sr.zenith, JD, sr.latitude, sr.longitude))
                sr.sunriseLocal = sr.sunriseUTC.AddHours(-sr.timezone + sr.dst)
                ' round to nearest minute.
                sr.sunriseLocal = IIf(sr.sunriseLocal.Second >= 30, sr.sunriseLocal.AddMinutes(1).AddSeconds(-sr.sunriseLocal.Second).AddMilliseconds(-sr.sunriseLocal.Millisecond), sr.sunriseLocal.AddSeconds(-sr.sunriseLocal.Second).AddMilliseconds(-sr.sunriseLocal.Millisecond))    'TODO: For performance reasons this should be changed to nested IF statements
            End If

            If nosunset = True Then
                ' there is no sunset
                ' if Northern hemisphere and spring or summer, OR
                ' if Southern hemisphere and fall or winter, use
                ' previous sunrise and next sunset
                If sr.latitude > 66.4 AndAlso sr.doy > 79 AndAlso sr.doy < 267 OrElse (sr.latitude < -66.4 AndAlso (sr.doy < 83 OrElse sr.doy > 263)) Then
                    Dim newjd As Double = findNextSunset(JD, sr.latitude, sr.longitude, sr.zenithvalue, sr)

                    Dim newsunset As DateTime = calcDateFromJD(newjd)
                    sr.sunsetUTC = newsunset.AddMinutes(calcSunsetUTC(sr.zenith, newjd, sr.latitude, sr.longitude))
                    sr.sunsetLocal = sr.sunsetUTC.AddHours(-sr.timezone + sr.dst)
                    ' round to nearest minute.
                    sr.sunsetLocal = IIf(sr.sunsetLocal.Second >= 30, sr.sunsetLocal.AddMinutes(1).AddSeconds(-sr.sunsetLocal.Second).AddMilliseconds(-sr.sunsetLocal.Millisecond), sr.sunsetLocal.AddSeconds(-sr.sunsetLocal.Second).AddMilliseconds(-sr.sunsetLocal.Millisecond))     'TODO: For performance reasons this should be changed to nested IF statements

                    ' if Northern hemisphere and fall or winter, OR
                    ' if Southern hemisphere and spring or summer, use
                    ' next sunrise and last sunset
                ElseIf sr.latitude > 66.4 AndAlso (sr.doy < 83 OrElse sr.doy > 263) OrElse (sr.latitude < -66.4 AndAlso sr.doy > 79 AndAlso sr.doy < 267) Then
                    Dim newjd As Double = findRecentSunset(JD, sr.latitude, sr.longitude, sr.zenithvalue, sr)

                    Dim newsunset As DateTime = calcDateFromJD(newjd)
                    sr.sunsetUTC = newsunset.AddMinutes(calcSunsetUTC(sr.zenith, newjd, sr.latitude, sr.longitude))
                    Dim sunsetLocal As DateTime = sr.sunsetUTC.AddHours(-sr.timezone + sr.dst)
                    ' round to nearest minute.
                    sr.sunsetLocal = IIf(sr.sunsetLocal.Second >= 30, sr.sunsetLocal.AddMinutes(1).AddSeconds(-sr.sunsetLocal.Second).AddMilliseconds(-sr.sunsetLocal.Millisecond), sr.sunsetLocal.AddSeconds(-sr.sunsetLocal.Second).AddMilliseconds(-sr.sunsetLocal.Millisecond))     'TODO: For performance reasons this should be changed to nested IF statements
                End If

            Else ' there is a sunset
                sr.sunsetUTC = sr.nowtimeZ.AddMinutes(calcSunsetUTC(sr.zenith, JD, sr.latitude, sr.longitude))
                sr.sunsetLocal = sr.sunsetUTC.AddHours(-sr.timezone + sr.dst)
                ' round to nearest minute.
                sr.sunsetLocal = IIf(sr.sunsetLocal.Second >= 30, sr.sunsetLocal.AddMinutes(1).AddSeconds(-sr.sunsetLocal.Second).AddMilliseconds(-sr.sunsetLocal.Millisecond), sr.sunsetLocal.AddSeconds(-sr.sunsetLocal.Second).AddMilliseconds(-sr.sunsetLocal.Millisecond))
            End If
        End If
    End Function 'doCalc 

    '/ <summary>
    '/ Converts radians to degrees
    '/ </summary>
    '/ <param name="angleRad">radians</param>
    Private Shared Function radToDeg(ByVal angleRad As Double) As Double
        Return 180.0 * angleRad / Math.PI
    End Function 'radToDeg


    '/ <summary>
    '/ Converts degrees to radians
    '/ </summary>
    '/ <param name="angleDeg">degrees</param>
    Private Shared Function degToRad(ByVal angleDeg As Double) As Double
        Return Math.PI * angleDeg / 180.0
    End Function 'degToRad


    '/ <summary>
    '/ Calculates the Gregorian DateTime from a Julian Day.
    '/ </summary>
    '/ <param name="jd">Julian Day</param>
    '/ <returns>DateTime in Gregorian calendar</returns>
    Private Shared Function calcDateFromJD(ByVal jd As Double) As DateTime
        Dim z As Double = Math.Floor(jd + 0.5)
        Dim f As Double = jd + 0.5 - z
        Dim A As Double
        If z < 2299161 Then
            A = z
        Else
            Dim alpha As Double = Math.Floor((z - 1867216.25) / 36524.25)
            A = z + 1 + alpha - Math.Floor(alpha / 4)
        End If
        Dim B As Double = A + 1524
        Dim C As Double = Math.Floor((B - 122.1) / 365.25)
        Dim D As Double = Math.Floor(365.25 * C)
        Dim E As Double = Math.Floor((B - D) / 30.6001)
        Dim day As Double = B - D - Math.Floor(30.6001 * E) + f
        Dim month As Double = IIf(E < 14, E - 1, E - 13) 'TODO: For performance reasons this should be changed to nested IF statements
        Dim year As Double = IIf(month > 2, C - 4716, C - 4715) 'TODO: For performance reasons this should be changed to nested IF statements
        Dim format As New System.Globalization.CultureInfo("en-US", True)

        Dim GregorianDate As DateTime = DateTime.Parse(month.ToString() & "/" & day.ToString() & "/" & year.ToString(), format)
        GregorianDate = GregorianDate.Date

        Return GregorianDate
    End Function 'calcDateFromJD


    '/ <summary>
    '/ Calculates the Julian Day corresponding to the date.
    '/ </summary>
    '/ <param name="day">the day</param>
    '/ <param name="month">the month</param>
    '/ <param name="year">the year</param>
    Private Shared Function calcJD(ByVal day As Integer, ByVal month As Integer, ByVal year As Integer) As Double

        '/ <returns>The Julian day corresponding to the date</returns>
        If month <= 2 Then
            year -= 1
            month += 12
        End If
        Dim A As Double = Math.Floor(year / 100)
        Dim B As Double = 2 - A + Math.Floor(A / 4)

        Dim JD As Double = Math.Floor(365.25 * (year + 4716)) + Math.Floor(30.6001 * (month + 1)) + day + B - 1524.5
        Return JD
    End Function 'calcJD

    '/ <summary>
    '/ Convert Julian Day to centuries since J2000.0.
    '/ </summary>
    '/ <param name="jd">the Julian Day to convert</param>
    Private Shared Function calcTimeJulianCent(ByVal jd As Double) As Double

        '/ <returns>returns the T value corresponding to the Julian Day</returns>
        Dim T As Double = (jd - 2451545.0) / 36525.0
        Return T
    End Function 'calcTimeJulianCent


    '/ <summary>
    '/ Convert centuries since J2000.0 to Julian Day.
    '/ </summary>
    '/ <param name="t">number of Julian centuries since J2000.0</param>
    Private Shared Function calcJDFromJulianCent(ByVal t As Double) As Double

        '/ <returns>returns the Julian Day corresponding to the t value</returns>
        Dim JD As Double = t * 36525.0 + 2451545.0
        Return JD
    End Function 'calcJDFromJulianCent


    '/ <summary>
    '/ Calculate the Geometric Mean Longitude of the Sun
    '/ </summary>
    '/ <param name="t">number of Julian centuries since J2000.0</param>
    Private Shared Function calcGeomMeanLongSun(ByVal t As Double) As Double

        '/ <returns>returns the Geometric Mean Longitude of the Sun in degrees</returns>
        Dim L0 As Double = 280.46646 + t * (36000.76983 + 0.0003032 * t)
        While L0 > 360.0
            L0 -= 360.0
        End While
        While L0 < 0.0
            L0 += 360.0
        End While
        Return L0 ' in degrees
    End Function 'calcGeomMeanLongSun


    '/ <summary>
    '/ Calculate the Geometric Mean Anomaly of the Sun
    '/ </summary>
    '/ <param name="t">number of Julian centuries since J2000.0</param>
    Private Shared Function calcGeomAnomalySun(ByVal t As Double) As Double

        '/ <returns>returns the Geometric Mean Anomaly of the Sun in degrees</returns>
        Dim M As Double = 357.52911 + t * (35999.05029 - 0.0001537 * t)
        Return M ' in degrees
    End Function 'calcGeomAnomalySun


    '/ <summary>
    '/ Calculate the equation of center for the sun
    '/ </summary>
    '/ <param name="t">number of Julian centuries since J2000.0</param>
    Private Shared Function calcSunEqOfCenter(ByVal t As Double) As Double

        '/ <returns>return value in degrees</returns>
        Dim m As Double = calcGeomAnomalySun(t)

        Dim mrad As Double = degToRad(m)
        Dim sinm As Double = Math.Sin(mrad)
        Dim sin2m As Double = Math.Sin(mrad + mrad)
        Dim sin3m As Double = Math.Sin(mrad + mrad + mrad)

        Dim C As Double = sinm * (1.914602 - t * (0.004817 + 0.000014 * t)) + sin2m * (0.019993 - 0.000101 * t) + sin3m * 0.000289
        Return C ' in degrees
    End Function 'calcSunEqOfCenter


    '/ <summary>
    '/ Calculate the true longitude of the sun
    '/ </summary>
    '/ <param name="t">number of Julian centuries since J2000.0</param>
    Private Shared Function calcSunTrueLong(ByVal t As Double) As Double

        '/ <returns>return the sun's true longitude in degrees</returns>
        Dim l0 As Double = calcGeomMeanLongSun(t)
        Dim c As Double = calcSunEqOfCenter(t)

        Dim O As Double = l0 + c
        Return O ' in degrees
    End Function 'calcSunTrueLong


    '/ <summary>
    '/ Calculate the true anamoly of the sun
    '/ </summary>
    '/ <param name="t">number of Julian centuries since J2000.0</param>
    Private Shared Function calcSunTrueAnomaly(ByVal t As Double) As Double

        '/<returns>the sun's true anamoly in degrees</returns>
        Dim m As Double = calcGeomAnomalySun(t)
        Dim c As Double = calcSunEqOfCenter(t)

        Dim v As Double = m + c
        Return v ' in degrees
    End Function 'calcSunTrueAnomaly


    '/ <summary>
    '/ Calculate the distance to the sun in AU
    '/ </summary>
    '/ <param name="t">number of Julian centuries since J2000.0</param>
    Private Shared Function calcSunRadVector(ByVal t As Double) As Double

        '/<returns>the sun radius vector in AUs</returns>
        Dim v As Double = calcSunTrueAnomaly(t)
        Dim e As Double = calcEccentricityEarthOrbit(t)

        Dim R As Double = 1.000001018 * (1 - e * e) / (1 + e * Math.Cos(degToRad(v)))
        Return R ' in AUs
    End Function 'calcSunRadVector


    '/ <summary>
    '/ Calculate the eccentricity of earth's orbit
    '/ </summary>
    '/ <param name="t">number of Julian centuries since J2000.0</param>
    Private Shared Function calcEccentricityEarthOrbit(ByVal t As Double) As Double

        '/<returns>the unitless eccentricity</returns>
        Dim e As Double = 0.016708634 - t * (0.000042037 + 0.0000001267 * t)
        Return e ' unitless
    End Function 'calcEccentricityEarthOrbit


    '/ <summary>
    '/ Calculate the apparent longitude of the sun
    '/ </summary>
    '/ <param name="t">number of Julian centuries since J2000.0</param>
    Private Shared Function calcSunApparentLong(ByVal t As Double) As Double

        '/<returns>the sun's apparent longitude in degrees</returns>
        Dim o As Double = calcSunTrueLong(t)

        Dim omega As Double = 125.04 - 1934.136 * t
        Dim lambda As Double = o - 0.00569 - 0.00478 * Math.Sin(degToRad(omega))
        Return lambda ' in degrees
    End Function 'calcSunApparentLong


    '/ <summary>
    '/ Calculate the mean obliquity of the ecliptic
    '/ </summary>
    '/ <param name="t">number of Julian centuries since J2000.0</param>
    Private Shared Function calcMeanObliquityOfEcliptic(ByVal t As Double) As Double

        '/<returns>mean obliquity in degrees</returns>
        Dim seconds As Double = 21.448 - t * (46.815 + t * (0.00059 - t * 0.001813))
        Dim e0 As Double = 23.0 + (26.0 + seconds / 60.0) / 60.0
        Return e0 ' in degrees
    End Function 'calcMeanObliquityOfEcliptic


    '/ <summary>
    '/ Calculate the corrected obliquity of the ecliptic
    '/ </summary>
    '/ <param name="t">number of Julian centuries since J2000.0</param>
    Private Shared Function calcObliquityCorrection(ByVal t As Double) As Double

        '/<returns>corrected obliquity in degrees</returns>
        Dim e0 As Double = calcMeanObliquityOfEcliptic(t)

        Dim omega As Double = 125.04 - 1934.136 * t
        Dim e As Double = e0 + 0.00256 * Math.Cos(degToRad(omega))
        Return e ' in degrees
    End Function 'calcObliquityCorrection


    '/ <summary>
    '/ Calculate the right ascension of the sun
    '/ </summary>
    '/ <param name="t">number of Julian centuries since J2000.0</param>
    Private Shared Function calcSunRtAscension(ByVal t As Double) As Double

        '/<returns>the sun's right ascension in degrees</returns>
        Dim e As Double = calcObliquityCorrection(t)
        Dim lambda As Double = calcSunApparentLong(t)

        Dim tananum As Double = Math.Cos(degToRad(e)) * Math.Sin(degToRad(lambda))
        Dim tanadenom As Double = Math.Cos(degToRad(lambda))
        Dim alpha As Double = radToDeg(Math.Atan2(tananum, tanadenom))
        Return alpha ' in degrees
    End Function 'calcSunRtAscension


    '/ <summary>
    '/ Calculate the declination of the sun
    '/ </summary>
    '/ <param name="t">number of Julian centuries since J2000.0</param>
    Private Shared Function calcSunDeclination(ByVal t As Double) As Double

        '/<returns>the sun's declination in degrees</returns>
        Dim e As Double = calcObliquityCorrection(t)
        Dim lambda As Double = calcSunApparentLong(t)

        Dim sint As Double = Math.Sin(degToRad(e)) * Math.Sin(degToRad(lambda))
        Dim theta As Double = radToDeg(Math.Asin(sint))
        Return theta ' in degrees
    End Function 'calcSunDeclination


    '/ <summary>
    '/ Calculate the difference between true solar time and mean solar time
    '/ </summary>
    '/ <param name="t">number of Julian centuries since J2000.0</param>
    Private Shared Function calcEquationOfTime(ByVal t As Double) As Double

        '/<returns>equation of time in minutes of time</returns>
        Dim epsilon As Double = calcObliquityCorrection(t)
        Dim l0 As Double = calcGeomMeanLongSun(t)
        Dim e As Double = calcEccentricityEarthOrbit(t)
        Dim m As Double = calcGeomAnomalySun(t)

        Dim y As Double = Math.Tan(degToRad(epsilon) / 2.0)
        y *= y

        Dim sin2l0 As Double = Math.Sin(2.0 * degToRad(l0))
        Dim sinm As Double = Math.Sin(degToRad(m))
        Dim cos2l0 As Double = Math.Cos(2.0 * degToRad(l0))
        Dim sin4l0 As Double = Math.Sin(4.0 * degToRad(l0))
        Dim sin2m As Double = Math.Sin(2.0 * degToRad(m))

        Dim Etime As Double = y * sin2l0 - 2.0 * e * sinm + 4.0 * e * y * sinm * cos2l0 - 0.5 * y * y * sin4l0 - 1.25 * e * e * sin2m

        Return radToDeg(Etime) * 4.0 ' in minutes of time
    End Function 'calcEquationOfTime


    '/ <summary>
    '/ Calculate the hour angle of the sun at sunrise for the latitude
    '/ </summary>
    '/ <param name="zenith">horizon limit</param>
    '/ <param name="latitude">latitude of observer in degrees</param>
    '/ <param name="solarDec">declination angle of sun in degrees</param>
    Private Shared Function calcHourAngleSunrise(ByVal zenith As Double, ByVal latitude As Double, ByVal solarDec As Double) As Double

        '/<returns>hour angle of sunrise in radians</returns>
        Dim latRad As Double = degToRad(latitude)
        Dim sdRad As Double = degToRad(solarDec)
        Dim HAarg As Double = Math.Cos(degToRad(zenith)) / (Math.Cos(latRad) * Math.Cos(sdRad)) - Math.Tan(latRad) * Math.Tan(sdRad)
        Dim HA As Double = Math.Acos(Math.Cos(degToRad(zenith)) / (Math.Cos(latRad) * Math.Cos(sdRad)) - Math.Tan(latRad) * Math.Tan(sdRad))
        Return HA ' in radians
    End Function 'calcHourAngleSunrise


    '/ <summary>
    '/ Calculate the hour angle of the sun at sunset for the latitude
    '/ </summary>
    '/ <param name="zenith">horizon limit</param>
    '/ <param name="latitude">latitude of observer in degrees</param>
    '/ <param name="solarDec">declination angle of sun in degrees</param>
    Private Shared Function calcHourAngleSunset(ByVal zenith As Double, ByVal latitude As Double, ByVal solarDec As Double) As Double

        '/<returns>hour angle of sunset in radians</returns>
        Dim latRad As Double = degToRad(latitude)
        Dim sdRad As Double = degToRad(solarDec)
        Dim HAarg As Double = Math.Cos(degToRad(zenith)) / (Math.Cos(latRad) * Math.Cos(sdRad)) - Math.Tan(latRad) * Math.Tan(sdRad)
        Dim HA As Double = Math.Acos(Math.Cos(degToRad(zenith)) / (Math.Cos(latRad) * Math.Cos(sdRad)) - Math.Tan(latRad) * Math.Tan(sdRad))
        Return -HA  ' in radians
    End Function 'calcHourAngleSunset


    '/ <summary>
    '/ Calculate the Universal Coordinated Time (UTC) of sunrise for the given day at
    '/ the given location on earth
    '/ </summary>
    '/ <param name="zenith">horizon limit</param>
    '/ <param name="JD">julian day</param>
    '/ <param name="latitude">latitude of observer in degrees</param>
    '/ <param name="longitude">longitude of observer in degrees</param>
    Private Shared Function calcSunriseUTC(ByVal zenith As Double, ByVal JD As Double, ByVal latitude As Double, ByVal longitude As Double) As Double

        '/<returns>time in minutes from zero Z</returns>
        Dim t As Double = calcTimeJulianCent(JD)

        ' *** Find the time of solar noon at the location, and use
        '     that declination. This is better than start of the
        '     Julian day
        Dim noonmin As Double = calcSolNoonUTC(t, longitude)
        Dim tnoon As Double = calcTimeJulianCent((JD + noonmin / 1440.0))

        ' *** First pass to approximate sunrise (using solar noon)
        Dim eqTime As Double = calcEquationOfTime(tnoon)
        Dim solarDec As Double = calcSunDeclination(tnoon)
        Dim hourAngle As Double = calcHourAngleSunrise(zenith, latitude, solarDec)

        Dim delta As Double = longitude - radToDeg(hourAngle)
        Dim timeDiff As Double = 4 * delta ' in minutes of time
        Dim timeUTC As Double = 720 + timeDiff - eqTime ' in minutes
        ' *** Second pass includes fractional jday in gamma calc
        Dim newt As Double = calcTimeJulianCent((calcJDFromJulianCent(t) + timeUTC / 1440.0))
        eqTime = calcEquationOfTime(newt)
        solarDec = calcSunDeclination(newt)
        hourAngle = calcHourAngleSunrise(zenith, latitude, solarDec)
        delta = longitude - radToDeg(hourAngle)
        timeDiff = 4 * delta
        timeUTC = 720 + timeDiff - eqTime ' in minutes
        Return timeUTC
    End Function 'calcSunriseUTC


    '/ <summary>
    '/ Calculate the Universal Coordinated Time (UTC) of solar noon for the given day
    '/ at the given location on earth
    '/ </summary>
    '/ <param name="t">number of Julian centuries since J2000.0</param>
    '/ <param name="longitude">longitude of observer in degrees</param>
    Private Shared Function calcSolNoonUTC(ByVal t As Double, ByVal longitude As Double) As Double

        '/<returns>time in minutes from zero Z</returns>
        Dim newt As Double = calcTimeJulianCent((calcJDFromJulianCent(t) + 0.5 + longitude / 360.0))
        Dim eqTime As Double = calcEquationOfTime(newt)
        Dim solarNoonDec As Double = calcSunDeclination(newt)
        Dim solNoonUTC As Double = 720 + longitude * 4 - eqTime ' min
        Return solNoonUTC
    End Function 'calcSolNoonUTC


    '/ <summary>
    '/ Calculate the Universal Coordinated Time (UTC) of sunset for the given day at
    '/ the given location on earth
    '/ </summary>
    '/ <param name="zenith">horizon limit</param>
    '/ <param name="JD">julian day</param>
    '/ <param name="latitude">latitude of observer in degrees</param>
    '/ <param name="longitude">longitude of observer in degrees</param>
    Private Shared Function calcSunsetUTC(ByVal zenith As Double, ByVal JD As Double, ByVal latitude As Double, ByVal longitude As Double) As Double

        '/<returns>time in minutes from zero Z</returns>
        Dim t As Double = calcTimeJulianCent(JD)

        ' *** Find the time of solar noon at the location, and use
        '     that declination. This is better than start of the
        '     Julian day
        Dim noonmin As Double = calcSolNoonUTC(t, longitude)
        Dim tnoon As Double = calcTimeJulianCent((JD + noonmin / 1440.0))

        ' First calculates sunrise and approx length of day
        Dim eqTime As Double = calcEquationOfTime(tnoon)
        Dim solarDec As Double = calcSunDeclination(tnoon)
        Dim hourAngle As Double = calcHourAngleSunset(zenith, latitude, solarDec)

        Dim delta As Double = longitude - radToDeg(hourAngle)
        Dim timeDiff As Double = 4 * delta
        Dim timeUTC As Double = 720 + timeDiff - eqTime

        ' first pass used to include fractional day in gamma calc
        Dim newt As Double = calcTimeJulianCent((calcJDFromJulianCent(t) + timeUTC / 1440.0))
        eqTime = calcEquationOfTime(newt)
        solarDec = calcSunDeclination(newt)
        hourAngle = calcHourAngleSunset(zenith, latitude, solarDec)

        delta = longitude - radToDeg(hourAngle)
        timeDiff = 4 * delta
        timeUTC = 720 + timeDiff - eqTime ' in minutes
        Return timeUTC
    End Function 'calcSunsetUTC


    '/ <summary>
    '/ Calculate the julian day of the most recent sunrise
    '/ </summary>
    '/ <param name='JD'>Julian Day</param>
    '/ <param name='latitude'>latitude of observer in degrees</param>
    '/ <param name='longitude'>longitude of observer in degrees</param>
    '/ <param name='zenithValue'>Official, Civil, Nautical or Astronimical horizon limit</param>
    Public Function findRecentSunrise(ByVal JD As Double, ByVal latitude As Double, ByVal longitude As Double, ByVal zenithValue As String, ByRef sr As calculations) As Double
        '      Dim sr As calculations
        '/ <returns>julian day of the most recent sunset</returns>
        Select Case zenithValue
            Case "O"
                sr.zenith = 90.8333
            Case "C"
                sr.zenith = 96.0
            Case "N"
                sr.zenith = 102.0
            Case "A"
                sr.zenith = 108.0
        End Select

        Dim julianday As Double = JD

        Dim time As Double = calcSunriseUTC(sr.zenith, JD, latitude, longitude)
        While [Double].IsNaN(time)
            julianday -= 1.0
            time = calcSunriseUTC(sr.zenith, julianday, latitude, longitude)
        End While
        Return julianday
    End Function 'findRecentSunrise


    '/ <summary>
    '/ Calculate the julian day of the most recent sunset
    '/ </summary>
    '/ <param name='JD'>Julian Day</param>
    '/ <param name='latitude'>latitude of observer in degrees</param>
    '/ <param name='longitude'>longitude of observer in degrees</param>
    '/ <param name='zenithValue'>Official, Civil, Nautical or Astronimical horizon limit</param>
    Public Function findRecentSunset(ByVal JD As Double, ByVal latitude As Double, ByVal longitude As Double, ByVal zenithValue As String, ByRef sr As calculations) As Double
        '  Dim sr As calculations
        '/ <returns>julian day of the most recent sunset</returns>
        Select Case zenithValue
            Case "O"
                sr.zenith = 90.8333
            Case "C"
                sr.zenith = 96.0
            Case "N"
                sr.zenith = 102.0
            Case "A"
                sr.zenith = 108.0
        End Select

        Dim julianday As Double = JD

        Dim time As Double = calcSunsetUTC(sr.zenith, JD, latitude, longitude)
        While [Double].IsNaN(time)
            julianday -= 1.0
            time = calcSunsetUTC(sr.zenith, julianday, latitude, longitude)
        End While
        Return julianday
    End Function 'findRecentSunset


    '/ <summary>
    '/ Calculate the julian day of the next sunrise
    '/ starting from the given day at the given location on earth
    '/ </summary>
    '/ <param name='JD'>Julian Day</param>
    '/ <param name='latitude'>latitude of the observer in degrees</param>
    '/ <param name='longitude'>longitude of the observer in degrees</param>
    '/ <param name='zenithValue'>Official, Civil, Nautical or Astronimical horizon limit</param>
    Public Function findNextSunrise(ByVal JD As Double, ByVal latitude As Double, ByVal longitude As Double, ByVal zenithValue As String, ByRef sr As calculations) As Double
        ' Dim sr As calculations
        '/ <returns>julian day of the next sunrise</returns>
        Select Case zenithValue
            Case "O"
                sr.zenith = 90.8333
            Case "C"
                sr.zenith = 96.0
            Case "N"
                sr.zenith = 102.0
            Case "A"
                sr.zenith = 108.0
        End Select

        Dim julianday As Double = JD

        Dim time As Double = calcSunriseUTC(sr.zenith, JD, latitude, longitude)
        While [Double].IsNaN(time)
            julianday += 1.0
            time = calcSunriseUTC(sr.zenith, julianday, latitude, longitude)
        End While
        Return julianday
    End Function 'findNextSunrise


    '/ <summary>
    '/ Calculate the julian day of the next sunset
    '/ starting from the given day at the given location on earth
    '/ </summary>
    '/ <param name='JD'>Julian Day</param>
    '/ <param name='latitude'>latitude of observer in degrees</param>
    '/ <param name='longitude'>longitude of observer in degrees</param>
    '/ <param name='zenithValue'>Official, Civil, Nautical or Astronimical horizon limit</param>
    Private Function findNextSunset(ByVal JD As Double, ByVal latitude As Double, ByVal longitude As Double, ByVal zenithValue As String, ByRef sr As calculations) As Double
        '   Dim sr As calculations
        '/ <returns>julian day of the next sunset</returns>
        Select Case zenithValue
            Case "O"
                sr.zenith = 90.8333
            Case "C"
                sr.zenith = 96.0
            Case "N"
                sr.zenith = 102.0
            Case "A"
                sr.zenith = 108.0
        End Select

        Dim julianday As Double = JD

        Dim time As Double = calcSunsetUTC(sr.zenith, JD, latitude, longitude)
        While [Double].IsNaN(time)
            julianday += 1.0
            time = calcSunsetUTC(sr.zenith, julianday, latitude, longitude)
        End While
        Return julianday
    End Function 'findNextSunset

End Class
